<?php
/**
 * This class adds the required scripts in the front-end to enable alternative loading.
 *
 * @package    Nelio_AB_Testing
 * @subpackage Nelio_AB_Testing/public/helpers
 * @author     David Aguilera <david.aguilera@neliosoftware.com>
 * @since      5.0.0
 */

defined( 'ABSPATH' ) || exit;

/**
 * This class adds the required scripts in the front-end to enable alternative loading.
 */
class Nelio_AB_Testing_Alternative_Loader {

	protected static $instance;
	const MAX_NUMBER_OF_POSSIBLE_COMBINATIONS = 24;

	public static function instance() {

		if ( is_null( self::$instance ) ) {
			self::$instance = new self();
		}//end if

		return self::$instance;

	}//end instance()

	public function init() {

		add_action( 'nab_relevant_experiments_loaded', [ $this, 'add_alternative_loading_hooks' ] );
		add_action( 'wp_head', [ $this, 'add_alternative_loader_script' ], 1 );
		add_action( 'wp_head', [ $this, 'maybe_add_canonical_link' ] );
		add_action( 'wp_enqueue_scripts', [ $this, 'register_alternative_loader_script' ] );

	}//end init()

	public function register_alternative_loader_script() {
		nab_register_script_with_auto_deps( 'nelio-ab-testing-alternative-loader', 'alternative-loader', false );
	}//end register_alternative_loader_script()

	public function add_alternative_loader_script() {

		$num_of_alternatives = $this->get_number_of_alternatives();

		$runtime  = Nelio_AB_Testing_Runtime::instance();
		$settings = Nelio_AB_Testing_Settings::instance();

		$alternative_loader_settings = array(
			'gdprCookie'            => $runtime->get_gdpr_cookie(),
			'hideQueryArgs'         => $settings->get( 'hide_query_args' ),
			'ignoreTrailingSlash'   => $this->ignore_trailing_slash(),
			'whenIsGeoDataRequired' => $this->when_is_geo_data_required(),
			'isStagingSite'         => nab_is_staging(),
			'maxCombinations'       => self::MAX_NUMBER_OF_POSSIBLE_COMBINATIONS,
			'numOfAlternatives'     => $num_of_alternatives,
			'participationChance'   => $settings->get( 'percentage_of_tested_visitors' ),
			'referrerParam'         => $runtime->get_referrer_param(),
			'site'                  => nab_get_site_id(),
			'experiments'           => $this->get_simplified_experiments(),
			'isTestedPostRequest'   => $runtime->is_tested_post_request(),
		);

		$asset   = include nelioab()->plugin_path . '/assets/dist/js/alternative-loader.asset.php';
		$version = $asset['version'];
		printf(
			'<script type="text/javascript" src="%s" id="nelio-ab-testing-alternative-loader-js"></script>',
			esc_url( add_query_arg( 'version', $version, nelioab()->plugin_url . '/assets/dist/js/alternative-loader.js' ) )
		);
		echo "\n";

		echo '<script type="text/javascript" id="nelio-ab-testing-alternative-loader-js-after">/*nelio-ab-testing-alternative-loader*//* <![CDATA[ */';
		printf(
			'window.nabAddSingleAction( "alt-loader-init-available", function() { nabloader.init( %s ); } );',
			wp_json_encode( $alternative_loader_settings )
		);
		echo "/* ]]> */</script>\n";

	}//end add_alternative_loader_script()

	public function maybe_add_canonical_link() {

		$settings = Nelio_AB_Testing_Settings::instance();
		if ( ! $settings->get( 'add_canonical_link' ) ) {
			return;
		}//end if

		$runtime       = Nelio_AB_Testing_Runtime::instance();
		$requested_alt = $runtime->get_alternative_from_request();
		if ( false === $requested_alt ) {
			return;
		}//end if

		printf(
			"<link rel=\"canonical\" href=\"%s\">\n",
			esc_url( $runtime->get_untested_url() )
		);

	}//end maybe_add_canonical_link()

	public function add_alternative_loading_hooks( $experiments ) {

		$runtime       = Nelio_AB_Testing_Runtime::instance();
		$requested_alt = $runtime->get_alternative_from_request();
		if ( false === $requested_alt ) {
			return;
		}//end if

		if ( ! is_array( $experiments ) ) {
			$experiments = [ $experiments ];
		}//end if

		foreach ( $experiments as $experiment ) {

			$experiment_type = $experiment->get_type();

			$control      = $experiment->get_alternative( 'control' );
			$alternatives = $experiment->get_alternatives();
			$alternative  = $alternatives[ $requested_alt % count( $alternatives ) ];

			/**
			 * Fires when a certain alternative is about to be loaded as part of a split test.
			 *
			 * Use this action to add any hooks that your experiment type might require in order
			 * to properly load the alternative.
			 *
			 * @param array  $alternative    attributes of the active alternative.
			 * @param array  $control        attributes of the control version.
			 * @param int    $experiment_id  experiment ID.
			 * @param string $alternative_id alternative ID.
			 *
			 * @since 5.0.0
			 */
			do_action( "nab_{$experiment_type}_load_alternative", $alternative['attributes'], $control['attributes'], $experiment->get_id(), $alternative['id'] );

		}//end foreach

	}//end add_alternative_loading_hooks()

	private function get_number_of_alternatives() {

		$gcd = function( $n, $m ) use ( &$gcd ) {
			if ( 0 === $n || 0 === $m ) {
				return 1;
			}//end if
			if ( $n === $m && $n > 1 ) {
				return $n;
			}//end if
			return $m < $n ? $gcd( $n - $m, $n ) : $gcd( $n, $m - $n );
		};

		$lcm = function( $n, $m ) use ( &$gcd ) {
			return $m * ( $n / $gcd( $n, $m ) );
		};

		$experiments  = $this->get_experiments_that_load_alternative_content();
		$alternatives = array_unique(
			array_map(
				function( $experiment ) {
					return count( $experiment->get_alternatives() );
				},
				$experiments
			)
		);

		if ( empty( $alternatives ) ) {
			return 0;
		}//end if

		return array_reduce( $alternatives, $lcm, 1 );

	}//end get_number_of_alternatives()

	private function get_simplified_experiments() {

		return array_reduce(
			$this->get_experiments_that_load_alternative_content(),
			function ( $result, $experiment ) {
				$result[ $experiment->get_id() ] = array(
					'segments' => $this->simplify_segments( $experiment->get_segments() ),
					'type'     => $experiment->get_type(),
				);
				return $result;
			},
			[]
		);

	}//end get_simplified_experiments()

	private function simplify_segments( $segments ) {

		$segments = array_map(
			function( $index ) use ( $segments ) {
				$segment = $segments[ $index ];
				return array(
					'id'                => $index,
					'segmentationRules' => $segment['segmentationRules'],
				);
			},
			array_keys( $segments )
		);

		foreach ( $segments as &$segment ) {
			foreach ( $segment['segmentationRules'] as &$rule ) {
				unset( $rule['id'] );
			}//end foreach
		}//end foreach

		return $segments;

	}//end simplify_segments()

	private function when_is_geo_data_required() {
		$experiments = $this->get_experiments_that_load_alternative_content();
		foreach ( $experiments as $experiment ) {
			if ( $this->contains_geo_segmentation_rules( $experiment ) ) {
				return 'now';
			}//end if
		}//end foreach

		$experiments = nab_get_running_experiments();
		foreach ( $experiments as $experiment ) {
			if ( $this->contains_geo_segmentation_rules( $experiment ) ) {
				return 'future';
			}//end if
		}//end foreach

		return 'never';
	}//end when_is_geo_data_required()

	private function contains_geo_segmentation_rules( $experiment ) {
		$segments = $experiment->get_segments();
		foreach ( $segments as $segment ) {
			$rules = $segment['segmentationRules'];
			foreach ( $rules as $rule ) {

				if ( in_array( $rule['type'], [ 'nab/ip-address', 'nab/location' ] ) ) {
					return true;
				}//end if

			}//end foreach
		}//end foreach

		return false;
	}//end contains_geo_segmentation_rules()

	private function get_experiments_that_load_alternative_content() {

		$runtime     = Nelio_AB_Testing_Runtime::instance();
		$experiments = $runtime->get_relevant_running_experiments();

		return array_values(
			array_filter(
				$experiments,
				function( $experiment ) {

					$control         = $experiment->get_alternative( 'control' );
					$experiment_id   = $experiment->get_id();
					$experiment_type = $experiment->get_type();

					/**
					 * Whether the experiment should be excluded from adding a `nab` query arg in the current request or not.
					 *
					 * @param boolean $is_excluded   whether the experiment should be excluded from the current request or not.
					 *                               Default: `false`.
					 * @param array   $control       original version.
					 * @param int     $experiment_id id of the experiment.
					 *
					 * @since 5.0.4
					 */
					return ! apply_filters( "nab_{$experiment_type}_exclude_experiment_from_loading", false, $control['attributes'], $experiment_id );

				}
			)
		);

	}//end get_experiments_that_load_alternative_content()

	private function ignore_trailing_slash() {
		/**
		 * Filters whether alternative content loading should ignore the trailing slash in a URL when comparing the current URL and the URL of the alternative the visitor is supposed to see.
		 *
		 * If it’s set to ignore, `https://example.com/some-page` and `https://example.com/some-page/` will be considered the same page. Otherwise, they’ll be different.
		 *
		 * @param boolean $ignore_trailing_slash whether to ignore the trailing slash or not.
		 *
		 * @since 5.0.8
		 */
		return apply_filters( 'nab_ignore_trailing_slash_in_alternative_loading', true );
	}//end ignore_trailing_slash()

}//end class
